/******************************************************************************
*
*  FILENAME :	pattern.c
*
*  DESCRIPTION :
*
*	Generic pattern-matching package.
*
*  PUBLIC FUNCTIONS :
*
*	int pat_match(char *string, char *pattern)
*
*  NOTES :
*
*	Written by Paul Chesson (1994)
*	E-Mail: chesson@mundil.cs.mu.OZ.AU
*
*/


/******************************************************************************
*  INCLUDE FILES
******************************************************************************/

#include "bugbyte.h"


/******************************************************************************
*  PRIVATE DECLARATIONS
******************************************************************************/

static Boolean pat_match_pos(const char *string, const char *pattern,
			     char *str_pos, char *pat_pos);


/******************************************************************************
*  PUBLIC FUNCTIONS
******************************************************************************/

/******************************************************************************
*
*  NAME :	Boolean pat_match(char *string, char *pattern)
*
*  DESCRIPTION :
*
*	Determines if a string can be matched by a given pattern.
*
*  INPUTS :
*
*	PARAMETERS:	string		Exact string to be matched
*			pattern		Wildcard pattern
*		
*  OUTPUTS :
*
*	RETURN:		TRUE		If string matches pattern.
*			FALSE		Otherwise.
*
*  NOTES :
*
*	sh(1) style file matching metacharacters accepted:
*
*	*	zero or more of any characters
*	?	any single character
*	[...]	any single character inside the brackets
*		including subranges  eg. l-z
*	[!...]	any single character not inside the bracketed range
*	\	treats next character as literal, escaping metacharacters
*
*/

/*
 *	Make initial call to recursive pattern-matching function
 *	by setting the string and pattern positions to their start.
 */

Boolean pat_match(const char *string, const char *pattern)
{
    return pat_match_pos(string, pattern, (char *)string, (char *)pattern);
}


/******************************************************************************
*  PRIVATE FUNCTIONS
******************************************************************************/

/******************************************************************************
*
*  NAME :	Boolean pat_match_pos(const char *string, const char *pattern,
*				      char *str_pos, char *pat_pos)
*
*  DESCRIPTION :
*
*	Recursive pattern matching function.
*	Determines if a sub-string can be matched by a given sub-pattern.
*
*  INPUTS :
*
*	PARAMETERS:	string		Exact full string to be matched
*			pattern		Full wildcard pattern
*			str_pos		Pointer to sub-string
*			pat_pos		Pointer to sub-pattern
*		
*  OUTPUTS :
*
*	RETURN:		TRUE		If sub-string matches sub-pattern.
*			FALSE		Otherwise.
*
*/

static Boolean pat_match_pos(const char *string, const char *pattern,
			     char *str_pos, char *pat_pos)
{

    /* _____ Check for end of pattern or string _____ */

    if (*pat_pos == NUL)
	return (*str_pos == NUL);
	
    if (*str_pos == NUL)
    {
	while (*pat_pos == '*')
	    pat_pos++;
	return (*pat_pos == NUL);
    }

    switch (*pat_pos)
    {

	/*
	 *	? : Match any single character
	 *
	 *	* : Match zero or more of any charcaters
	 */

	case '?' :
	    return pat_match_pos(string, pattern, str_pos+1, pat_pos+1);

	case '*' :
	    return (pat_match_pos(string, pattern, str_pos+1, pat_pos+1)
		|| (pat_match_pos(string, pattern, str_pos+1, pat_pos)));
		
	/*
	 *	[...]  : Match one of the enclosed characters in set
	 *	[!...] : Match one character not in the set
	 *
	 *	a-k    : Matches sub-range of characters
	 *
	 *	!!! CAREFUL NOT TO INCREMENT pat_pos PAST END OF pattern !!!
	 */

	case '[' :
	{
	    Boolean	match = FALSE;		/* is next char in set?	*/
	    Boolean	comp_set = FALSE;	/* take set complement?	*/
	    char	low = NUL;		/* start of sub-range	*/

	    if (*(pat_pos+1) == '!')
	    {
		comp_set = TRUE;
		pat_pos++;
	    }

	    while (*++pat_pos != ']' && *pat_pos != NUL)
	    {
		/* _____ sub-range _____ */

		if (*pat_pos == '-' && low != NUL && *(pat_pos+1) != ']')
		{
		    if (*(pat_pos+1) == '\\' && *++pat_pos == NUL)
		        return FALSE;
		    if (*str_pos > low && *str_pos <= *(pat_pos+1))
			match = TRUE;
		    low = *(pat_pos+1);
		}

		/* _____ single character _____ */

		else
		{
		    if (*pat_pos == '\\' && *++pat_pos == NUL)
			return FALSE;
		    if (*pat_pos == *str_pos)
			match = TRUE;
		    low = *pat_pos;
		}

	    }  /* while still inside set */

	    if ( (match && !comp_set) || (!match && comp_set) )
	        pat_match_pos(string, pattern, str_pos+1, pat_pos+1);
	    else
		return FALSE;

	}  /* case '[' */

	/*
	 *	Single characters
	 *
	 *	\ : Interprets next character as literal, escaping metachars
	 */

	case '\\' :
	    if (*++pat_pos == NUL)
		return FALSE;		/* fall-through case intended	*/

	default :
	    if (*pat_pos != *str_pos)
		return FALSE;
	    return pat_match_pos(string, pattern, str_pos+1, pat_pos+1);

    }  /* pattern character switch */

}

